#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
#include <vector>
#include <sstream>
#include <iterator>
#include <algorithm>

#include <assert.h>

#define ASSERT(c) assert(c)

#include "node.h"
#include "nodelist.h"

using namespace std;



/* constructor */
nodelist::nodelist( string filestr ){

    /* get all node list */
    vector<int> fullnodelevel,   nodelevel;
    vector<string> fullnodelist, nodelist;

    setNodelist( filestr, fullnodelist );
    setNodelevel( fullnodelist, fullnodelevel );

    /* parsing check */
    /*
    for(int i=0; i<(int)fullnodelevel.size(); i++){ 
    	printf("%3d ", i); 
    	printf("%3d ", fullnodelevel[i]); 
	for(int j=0;j<fullnodelevel[i]; j++){ printf(" |"); }
    	cout << fullnodelist[i] << endl; 
    }
    cout << "=================" << endl;
    */
    
    /* remove close tag */
    nodesize = removeCloseTag( fullnodelevel, fullnodelist, nodelevel, nodelist );
    /*
    for(int i=0; i<(int)nodelevel.size(); i++){ 
    	printf("%3d ", i); 
    	printf("%3d ", nodelevel[i]); 
	for(int j=0;j<nodelevel[i]; j++){ printf(" |"); }
    	cout << nodelist[i] << endl; 
    }
    cout << "nodesize=" << nodesize << endl;
    */

    /* set node */
    nodes = new node[nodesize];
    for(int i=0; i<(int)nodelist.size(); i++){
	node temp;	
	temp.setID(i);
	temp.setLevel( nodelevel[i] );
	temp.setHtmlNode( nodelist[i] );
    
	temp.setParent( searchParent(i, nodelevel, nodes ) );
	searchChildren( i, nodelevel, nodes );

	nodes[i] = temp;
	
    }



}

nodelist::~nodelist(){
   delete[] nodes;
}    


/* set */
void nodelist::setNodelist( string str, vector<string>& tags ) {

    string  s, sa, sb, sl;
    vector<string> temp1; 
    vector<char> tagStack;

    /* split with tag */
    for(int i=0; i<(int)str.size(); i++){
	if( str[i] == '\n' || str[i] == '\t' || str[i] == '\0' || str[i] == '\r' ){ continue; }

	if( str[i] == '<' && tagStack.size() == 0 ){ 
	    sa = sa + "\n";
	}

	if( str[i] == '<' && tagStack.size() == 0 ||
	    ( i>3 && str[i-3] == '<' && str[i-2] == '!' && str[i-1] == '-' && str[i] == '-' ) ){
	    tagStack.push_back( str[i] );
	} else if( (tagStack.size() > 0 && str[i] == '>' ) ){ tagStack.pop_back(); } 
	
	/* comment end */
	if( (tagStack.size() > 0 && i>2 && str[i-2] == '-' && str[i-1] == '-' && str[i] == '>' ) ){ 
	    tagStack.pop_back(); 
	}

	sa =sa +  str[i] ;

	if( str[i] == '>' && tagStack.size() == 0 ){ 
	    sa = sa + "\n"; 
	}

    }

    /* remove space line */
    for(int i=1; i<(int)sa.size(); i++){
	if( sa[i] == '\n' ){
	    if( sa[i-1] == '\n'){ continue; }
	}
	sb = sb +  sa[i];
    }

    /* push to tags */
    int space=0;
    for(int i=0; i<(int)sb.size(); i++){
	if( sb[i] == '\n' ){
	    for(int j=0; j<(int)sl.size(); j++){
	    	if( sl[j] == ' ' ){ space++; }
	    }
	    if( space != (int)sl.size() ){
		tags.push_back( getLowerString(sl) ); 
		space = 0;
		sl.clear();
	    }
	    sl.clear(); 
	    space = 0;
	    continue;
	}
	sl = sl + sb[i];
    }

    return;
}

int nodelist::isIgnoreTag ( string& tagname ){
    // return value 1: match  -1: missmatch
    
    int i=0;
    char* ignoretags[] 
	= { "BR","br", "META", "meta", 
	    "link","LINK","img", "IMG",	
	    "input","INPUT","hr","HR",
	    "area","AREA","isimport","ISIMPORT",
	    "X-CLARIS","x-claris","col","COL",
	    "SPACER","spacer","BASEFONT","basefont",
	    "WBR","wbr","ah","AH",
	    "BASE","BASE", "!DOCTYPE", "!doctype",
	    "BGSOUND","bgsound","EMBED","embed",
	    "script","SCRIPT","ISINDEX","isindex",
	    "KEYGEN","keygen","NEXTID","nextid",
	    "?","/","!","!--",
	    "PARAM","param","PLAINTEXT","plaintext","-1",
	 };
    
    while( ignoretags[i]  != "-1" ){
	if( tagname == ignoretags[i] ){ return 1; }
	i++;
    } 

    return -1;
}

int nodelist::checkRule (string& before, string& after){
    int i=0;
    int value=0; // -1:error 0:ignore 1...4:pop

    char * rule[][3] ={ {"a","td","2"},
			{"a","/div","2"} ,
			{"a","/img","-1"},
			{"a","/span","2"},
			{"a","a","0"},
			{"a","table","1"},
			{"b","/font","2"},
			{"b","/tr","2"},
			{"b","/td","2"},
			{"B","P","1"}, {"b","p","1"},
			{"base","/head","2"},
			{"caption","tr","1"},
			{"caption","td","1"},
			{"csaction","csaction","1"},
			{"csactions","/head","2"},
			{"center","/font","1"},
			{"center","/td","2"},
			{"center","/tr","-1"},
			{"center","center","1"},
			{"center","/div","2"},
			{"center","/body","2"},
			{"dl","/div","2"},
			{"dl","/table","2"},
			{"dl","/td","1"},
			{"dd","dd","1"},
			{"dt","dt","1"},
			{"dd","/font","2"},
			{"dd","/dl","2"},
			{"dd","/td","3"},
			{"dt","/td","2"},
			{"dt","dd","1"},
			{"dt","/dl","2"},
			{"dt","dt","1"},
			{"div","/p","0"},
			{"div","/td","2"},
			{"div","/font","2"},
			{"div","/body","2"},
			{"div","/dl","2"},
			{"form","/td","2"},
			{"font","/a","2"},
			{"font","/b","2"},
			{"font","/td","2"},
			{"font","/div","2"},
			{"font","/center","2"},
			{"font","/table","2"},
			{"font","table","1"},
			{"font","p","2"},
			{"font","center","1"},
			{"li","ol","2"},
			{"li","/ul","2"},
			{"li","ul","2"},
			{"li","/body","3"},
			{"li","li","1"},
			{"li","/td","2"},
			{"li","td","-1"},
			{"li","tr","-1"},
			{"ul","tr","1"},
			{"ul","td","1"},
			{"rb","rb","1"},
			{"rb","/td","2"},
			{"rb","/ruby","2"},
			{"span","/td","2"},
			{"table","/font","-1"},
			{"table","/body","2"},
			{"td","td","1"},
			{"td","/table","3"},
			{"td","/tr","2"},
			{"td","tr","2"},
			{"td","/form","-1"},
			{"tr","/tr","1"},
			{"tr","/table","2"},
			{"tr","td","0"},
			{"tr","center","-1"},
			{"tr","/form","-1"},
			{"tr","/tbody","2"},
			{"td","/tbody","3"},
			{"td","/img","-1"},
			{"td","/body","4"},
			{"p","/body","2"},
			{"p","/center","2"},
			{"p","/font","1"},
			{"p","/form","2"},
			{"p","/div","2"}, 
			{"p","/dl","2"}, 
			{"p","/td","2"}, 
			{"p","/small","2"},
			{"p","td","2"},
			{"p","/td","1"},
			{"p","/tr","1"},
			{"p","/STRONG","2"},
			{"p","p","1"},
			{"p","/form","1"},
			{"p","/dd","2"},
			{"p","dt","2"},
			{"param","param","1"},
			{"param","object","2"},
			{"head","body","1"},
			{"body","/div","-1"},
			{"body","/td","-1"},
			{"body","/p","-1"},
			{"body","/form","-1"},
			{"body","/form","-1"},
			{"body","/tr","-1"},
			{"body","/table","-1"},
			{"body","/html","2"},
			{"body","body","1"},
			{"html","/div","-1"},
			{"html","/p","-1"},
			{"html","/form","-1"},
			{"html","/table","-1"},
			{"html","/link","-1"},
			{"html","/head","-1"},
			{"option","option","1"},
			{"option","/select","2"},
			{"-1","-1","-1"} //last
		    };

    while(1){
	if( rule[i][0] == "-1" ){ break; }

	if( rule[i][0] == before && rule[i][1] == after ){
	    value = atoi(rule[i][2]);
	    return value;
	}
	i++;
    }

    return value;
}

void nodelist::setNodelevel( vector<string>& nodelist, vector<int>& taglevel  ) {
    vector<string> tagStack;
    string lastTagname, nowTagname;;
    int level=0,rule=0;
    bool debugflag=false;  /* true:print false:don't print */

    /* init. stack */
    tagStack.push_back("");

    /* parse html tag for each node */
    for(int i=0; i<(int)nodelist.size(); i++){
	
	if( nodelist[i][0] == '<' && nodelist[i][1] == '!' && 
	    nodelist[i][2] == '-' && nodelist[i][3] == '-'		/* comment tag */
	){	   
	    taglevel.push_back(level);
	    if( debugflag == true ){ for(int j=0;j<level;j++){ cout << " "; } cout << nodelist[i] << endl; }

	} else if( nodelist[i][0] == '<' && nodelist[i][1] == '/'){	/* close tag */
	    nowTagname  = getTagname( nodelist[i] );
	    lastTagname = getTagname( tagStack.back() );

	    /* transform to small char. */
	    nowTagname = getLowerString( nowTagname );
	    lastTagname = getLowerString( lastTagname );
	    /*cout << lastTagname << " " << nowTagname << endl; */

	    /* check for rule */
    	    rule = checkRule(lastTagname,nowTagname); 
	    if( rule == -1 ){ 
		if( debugflag == true ){ cout << "close error!!" << endl; }
		if( debugflag == true ){ for(int j=0;j<level;j++){ cout << " "; } cout << nodelist[i] << endl; }
	    } else if( rule == 1 ){
		level--; taglevel.push_back(level);
		tagStack.pop_back();
		if( debugflag == true ){ for(int j=0;j<level;j++){ cout << " "; } cout << nodelist[i] << endl; }
		continue;
	    } else if( rule == 2 ){
		level = level -2; taglevel.push_back(level);
		tagStack.pop_back(); tagStack.pop_back();
		if( debugflag == true ){ for(int j=0;j<level;j++){ cout << " "; } cout << nodelist[i] << endl; }
		continue;
	    } else if( rule == 3 ){
		level = level -3; taglevel.push_back(level);
		tagStack.pop_back(); tagStack.pop_back(); tagStack.pop_back();
		if( debugflag == true ){ for(int j=0;j<level;j++){ cout << " "; } cout << nodelist[i] << endl; }
		continue;
	    } else if( rule == 4 ){
		level = level -4; taglevel.push_back(level);
		tagStack.pop_back(); tagStack.pop_back();
		tagStack.pop_back(); tagStack.pop_back();
		if( debugflag == true ){ for(int j=0;j<level;j++){ cout << " "; } cout << nodelist[i] << endl; }
		continue;
	    }	

	    /* match error */
	    lastTagname = "/" + lastTagname;
	    if( nowTagname != lastTagname ){
	    	taglevel.push_back(level);
		if( debugflag == true ){ cout << "error!!" << endl; }
	    	continue;
	    }
	    level--; 
	    tagStack.pop_back();
	    taglevel.push_back(level);

	    if( debugflag == true ){ for(int j=0;j<level;j++){ cout << " "; } cout << nodelist[i] << endl; }

	} else if( nodelist[i][0] == '<'  ){				/* open tag */
	    /* check for rule */
	    nowTagname = getTagname( nodelist[i] );
	    if( ! tagStack.empty() ){
		lastTagname = getTagname( tagStack.back() );
	    }

	    /* transform to small char. */
	    lastTagname = getLowerString( lastTagname );
	    nowTagname  = getLowerString( nowTagname  );

	    rule = checkRule(lastTagname,nowTagname);
	    if( rule == -1 ){
		cout << "open error" << endl;
	    } else if( rule == 1 ){
		level--; 
		tagStack.pop_back();
	    } else if( rule == 2 ){
		level = level-2; 
		tagStack.pop_back();
		tagStack.pop_back();
	    }
	    taglevel.push_back(level);
	    if( debugflag == true ){ for(int j=0;j<level;j++){ cout << " "; } cout << nodelist[i] << endl; }

	    if( isIgnoreTag( nowTagname ) != 1 ){
		tagStack.push_back( nodelist[i] );
		level++; 
	    }

	} else {							    /* text */
	    taglevel.push_back(level);
	    if( debugflag == true ){ for(int j=0;j<level;j++){ cout << " "; } cout << nodelist[i] << endl; }

	}
    }

    return; 
}

int nodelist::removeCloseTag( vector<int>& fullnodelevel,  vector<string>& fullnodelist,  /* input  */
			      vector<int>& nodelevel,      vector<string>& nodelist ){	 /* output */
    int size=0;
 
    for(int i=0; i<(int)fullnodelist.size(); i++){
	if( fullnodelist[i][0] == '<' && fullnodelist[i][1] == '/'){ continue; }
	if( fullnodelist[i][0] == '<' && fullnodelist[i][1] == '!'){ continue; }
	if( fullnodelist[i][0] == '<' && fullnodelist[i][1] == '!' &&
	    fullnodelist[i][2] == '-' && fullnodelist[i][3] == '-'  ){ continue; }
	nodelevel.push_back( fullnodelevel[i] );
	nodelist.push_back( fullnodelist[i] );
	size++;
    }

    return size;
} 

/*
void nodelist::removeCloseTag( vector<int>& level,  vector<string>& list ){	

    vector<int>::iterator ite2 = level.begin();
    for( vector<string>::iterator ite = list.begin(); ite != list.end(); ){
	if( ( (*ite)[0] == '<' && (*ite)[1] == '/')  ||
	    ( (*ite)[0] == '<' && (*ite)[1] == '!')  ||
	    ( (*ite)[0] == '<' && (*ite)[1] == '!' && (*ite)[2] == '-' && (*ite)[3] == '-'  ) 
	) {
	    ite = list.erase( ite );
	    ite2 = level.erase( ite2 );
	} else { ite++; ite2++; }
    }

    return;
} 
*/

node* nodelist::searchParent(int ID, vector<int>& level, node* array ){ 
    //cout << "ID=" << ID << endl;
    for(int i=ID; i>=0; i--){
	if( level[i] == (level[ID]-1) ){
	    return &array[i];
	}
    } 

    return NULL;
}

void nodelist::searchChildren(int ID, vector<int>& level, node* array ){ 

    for(int i=ID; i>=0; i--){
	if( level[i] == (level[ID]-1) ){
	    array[i].setChildren( &array[ID] );
	    return ;
	}
    } 

    return;
}


/* get */

int nodelist::getSize(){
    return nodesize;
}

string nodelist::getTagname (string& tag ){ 
    string tagname;

    if( tag.empty() ){	/* empty tag */
	return "-1"; 
    } else if ( tag[0] != '<' ){ /* not tag */
	return "-1"; 
    } else if ( tag.substr(0,4) == "<!--" ){ /* comment tag */
	return "-1"; 
    }
	
    for(int i=1; i<(int)tag.size(); i++){
	if( tag[i] == ' ' || tag[i] == '>' ){ break; }
	tagname = tagname + tag[i];
    }

    return tagname;
}

string nodelist::getLowerString( string target ){
    string low_target;
    char *ch, *low_ch;
    char a;

    ch  = new char[target.size()+1];
    low_ch = new char[target.size()+1];

    // check target
    for(int i=0; i<(int)target.size(); i++){
	a = target[i];	
	if( a < 0x20 || a > 0x7E ){
	    //printf("not ascii\t"); cout << target << endl;
	    return target;
	} 
    }

    int i;
    for(i=0; i<(int)target.size(); i++){
	ch[i] = target[i];	     /* string -> char */
	low_ch[i] = tolower(ch[i]); /* change lower char */
    }
    low_ch[i++] = '\0';

    /* new assign for low_target */
    low_target.assign(low_ch);

    return low_target;
}

node* nodelist::getNode( int i ){
    return &nodes[i];
}

int nodelist::getID( int i ){
    return nodes[i].getID();
}

int nodelist::getLevel( int i ){
    return nodes[i].getLevel();
}

string nodelist::getHtmlNode( int i ){
    return nodes[i].getHtmlNode();
}

vector<node*> nodelist::getChildren( int i ){
    return nodes[i].getChildren();
}

node* nodelist::getParent( int i ){
    return nodes[i].getParent();
}

vector<node*> nodelist::getPath(int i){
    vector<node*> path;
    node* start = &nodes[i];

    while( start->getID() != 0 ){
        /* cout << start->getParent()->getHtmlNode() << endl;  */
        path.push_back( start->getParent() );
        start = start->getParent(); /* renew start point */
    }

    return path;
}

void nodelist::DFS( int i ){
    cout << nodes[i].getHtmlNode() << endl;

    vector<node*> children = nodes[i].getChildren();

    for(int i=0; i<(int)children.size(); i++){
        if( children.size() == 0 ){ return; }
        DFS( children[i]->getID() );
    }

    return;
}

void nodelist::DFS( int i, vector<node*>& path ){
    /* cout << nodes[i].getHtmlNode() << endl; */
    path.push_back( &nodes[i] );

    vector<node*> children = nodes[i].getChildren();

    for(int i=0; i<(int)children.size(); i++){
        if( children.size() == 0 ){ return; }
        DFS( children[i]->getID(), path );
    }

    return;
}

void nodelist::BFS(int i){

    vector<node*> q;
    vector<node*> w;    
    vector<node*> children;
    
    /* push root to stack q */
    q.push_back( &nodes[i] );

    while(1){
	if( q.size() == 0 ){ break; }

	/* push children to stack w */
	for(int i=0; i<(int)q.size(); i++){
	    //cout << q[i]->getHtmlNode() << endl;
	    children = q[i]->getChildren();
	    for(int j=0; j<(int)children.size(); j++){
		//cout << "push: " << children[j]->getID() << endl;
		w.push_back( children[j] );
	    }
	}

	/* print */
	for(int i=0; i<(int)w.size(); i++){
	    cout << w[i]->getHtmlNode() << endl;
	}
    
	/* copy w to q */
	q.erase( q.begin(), q.end() );
	for(int i=0; i<(int)w.size(); i++){
	    q.push_back( w[i] );
	}
	
	/* clear */
	w.clear();
	children.clear();
    } 


}

void nodelist::BFS(int i, vector<node*>& path ){

    vector<node*> q;
    vector<node*> w;    
    vector<node*> children;
    
    /* push root to stack q */
    q.push_back( &nodes[i] );

    while(1){
	if( q.size() == 0 ){ break; }

	/* push children to stack w */
	for(int i=0; i<(int)q.size(); i++){
	    //cout << q[i]->getHtmlNode() << endl;
	    children = q[i]->getChildren();
	    for(int j=0; j<(int)children.size(); j++){
		//cout << "push: " << children[j]->getID() << endl;
		w.push_back( children[j] );
	    }
	}

	/* print */
	/*
	for(int i=0; i<(int)w.size(); i++){
	    cout << w[i]->getHtmlNode() << endl;
	}
	*/

	/* push */
	for(int i=0; i<(int)w.size(); i++){
	    path.push_back( w[i] );
	}
    
	/* copy w to q */
	q.erase( q.begin(), q.end() );
	for(int i=0; i<(int)w.size(); i++){
	    q.push_back( w[i] );
	}
	
	/* clear */
	w.clear();
	children.clear();
    } 


}

vector<node*> nodelist::getLevelNodes(int start, int target ){

    vector<node*> q;
    vector<node*> w;    
    vector<node*> children;
    vector<node*> path;
    int level=0;
    
    /* push root to stack q */
    q.push_back( &nodes[start] );

    while(1){
	if( q.size() == 0 ){ break; }

	/* push children to stack w */
	for(int i=0; i<(int)q.size(); i++){
	    //cout << q[i]->getHtmlNode() << endl;
	    children = q[i]->getChildren();
	    for(int j=0; j<(int)children.size(); j++){
		//cout << "push: " << children[j]->getID() << endl;
		w.push_back( children[j] );
	    }
	}

	/* print */
	/* cout << "level=" << level << "-------" << endl; */
	if( level == target ){
	    for(int i=0; i<(int)w.size(); i++){
		/*cout << w[i]->getHtmlNode() << endl; */
		path.push_back( w[i] );
	    }
	}
    
	/* copy w to q */
	q.erase( q.begin(), q.end() );
	for(int i=0; i<(int)w.size(); i++){
	    q.push_back( w[i] );
	}
	
	/* clear */
	w.clear();
	children.clear();
	level++;
    } 

    return path;
}

bool nodelist::existCloseTag( string tagname ) {

    int i=0;
    string ignoretags[] 
	= { "BR","br", "META", "meta", 
	    "link","LINK","img", "IMG",	
	    "input","INPUT","hr","HR",
	    "area","AREA","isimport","ISIMPORT",
	    "X-CLARIS","x-claris","col","COL",
	    "SPACER","spacer","BASEFONT","basefont",
	    "WBR","wbr","ah","AH",
	    "BASE","BASE", "!DOCTYPE", "!doctype",
	    "BGSOUND","bgsound","EMBED","embed",
	    "script","SCRIPT","ISINDEX","isindex",
	    "KEYGEN","keygen","NEXTID","nextid",
	    "?","/","!","!--",
	    "PARAM","param","PLAINTEXT","plaintext","-1",
	 };

    while( ignoretags[i]  != "-1" ){
	if( tagname == ignoretags[i] ){ return false; }
	i++;
    } 

    return true;
}


void nodelist::writeHtml(int i, ofstream& fout) {

    for(int j=0; j<(int)nodes[i].getLevel(); j++){
	/* cout << " "; */
	fout << " ";
    }
    /* cout << nodes[i].getHtmlNode() << endl; */
    fout << nodes[i].getHtmlNode() << endl;
    
    /* get close tag */
    string closeTag = "";
    string tmp = nodes[i].getHtmlNode();
    string tagname = getTagname( tmp );

    if( tagname != "" && existCloseTag( tagname ) && tagname != "-1" ){
	closeTag = "</" + tagname + ">";
	closeTags.push_back( closeTag );

	int nowLevel = nodes[i].getLevel();
	levelStack.push_back( nowLevel );
    }
    
    /* children */
    vector<node*> children = nodes[i].getChildren();
    for(int j=0; j<(int)children.size(); j++){
        if( children.size() == 0 ){ return; }
        writeHtml( children[j]->getID(), fout );
    }

    /* close tag */
    if( levelStack.back() >= nodes[i].getLevel() ){

	for(int j=0; j<(int)levelStack.back(); j++){ 
	    /* cout << " "; */
	    fout << " "; 
	}
	string outCloseTag = closeTags.back(); 
	/* cout << outCloseTag << endl; */
	fout << outCloseTag << endl;

	closeTags.pop_back();
	levelStack.pop_back();
    }

    return;
}




#if 0

int main(int argc, char* argv[]){
    
    if( argc != 2 ){
	cout << "Usage" << endl;
	exit(1);
    }

    char* filename = argv[1];

    string fileString;
    
    fileString = file2string( filename );

    //nodelist doc( filename );


}

#endif

